use std::{
    io::{self, ErrorKind},
    time::Duration,
};

use nats::{Connection, Handler};

use crate::chat_message::ChatMessage;

const SERVER: &str = "127.0.0.1";
const SUBJECT_MESSAGES: &str = "telestion.chat";
const SUBJECT_REQEUST_REPLY_MESSAGES: &str = "telestion.request-reply";

pub fn connect_to_nats(username: &str, password: &str) -> io::Result<Connection> {
    nats::Options::with_user_pass(username, password)
        .with_name("Chat Application")
        .connect(SERVER)
}

pub fn subscribe_to_chat_messages(username: String, nc: &Connection) -> io::Result<Handler> {
    // This runs the closure in a separate Thread
    let subscription =
        nc.subscribe(SUBJECT_MESSAGES)?
            .with_handler(move |msg| -> Result<(), io::Error> {
                match ChatMessage::try_from(msg.data) {
                    Ok(message) => {
                        // Do not show messages from me
                        if message.author != username {
                            println!("Received {}", message);
                        }
                    }
                    Err(e) => eprintln!("Error processing message: {}", e),
                };
                Ok(())
            });
    Ok(subscription)
}

pub fn subscribe_to_requests<F>(nc: &Connection, answer_request: F) -> io::Result<Handler>
where
    F: Fn(ChatMessage) -> ChatMessage + Send + 'static,
{
    // TODO: This is problematic due to the multithreading. Use multithreading safe things here!!!
    let nc = nc.clone();
    // This runs the closure in a separate Thread
    let subscription = nc.subscribe(SUBJECT_REQEUST_REPLY_MESSAGES)?.with_handler(
        move |msg| -> Result<(), io::Error> {
            let reply_address = msg
                .reply
                .expect("All messages on this channel have to contain a reply address!");
            match ChatMessage::try_from(msg.data) {
                Ok(message) => {
                    let answer = answer_request(message);
                    publish_message_to_channel(&reply_address, &answer, &nc)?;
                }
                Err(e) => eprintln!("Error processing message: {}", e),
            };
            Ok(())
        },
    );
    Ok(subscription)
}

pub fn publish_message(message: &ChatMessage, nc: &Connection) -> io::Result<()> {
    publish_message_to_channel(SUBJECT_MESSAGES, message, nc)
}

pub fn publish_invalid_message(nc: &Connection) -> io::Result<()> {
    nc.publish(SUBJECT_MESSAGES, "This is not a valid json")?;
    Ok(())
}

fn publish_message_to_channel(
    channel: &str,
    message: &ChatMessage,
    nc: &Connection,
) -> io::Result<()> {
    let json_string_utf8 = serde_json::to_string(message).expect("Cannot serialize message");
    // TODO: there is no error if server is offline
    nc.publish(channel, json_string_utf8)?;
    Ok(())
}

pub fn post_request(message: &ChatMessage, nc: &Connection) -> io::Result<()> {
    let json_string_utf8 = serde_json::to_string(message).expect("Cannot serialize message");
    let response = nc.request_timeout(
        SUBJECT_REQEUST_REPLY_MESSAGES,
        json_string_utf8,
        Duration::from_secs(5),
    );

    match response {
        Ok(response) => {
            match ChatMessage::try_from(response.data) {
                Ok(message) => {
                    println!("Received quickchat message: {}", message);
                }
                Err(e) => eprintln!("Error processing message: {}", e),
            };
            Ok(())
        }

        Err(e) => {
            if e.kind() == ErrorKind::NotFound {
                eprintln!("There is no one else online!");
                return Ok(());
            } else if e.kind() == ErrorKind::TimedOut {
                eprintln!("There was no fast enough response!");
                return Ok(());
            } else {
                return Err(e);
            }
        }
    }
}
